import click
import os
import os.path
import splogger as log
import spvm.core as core
import spvm.config as cfg


def get_project(projectname):
    if os.path.isabs(projectname):
        project = core.make_project_object(projectname)
    else:
        project = core.make_project_object(
            os.path.join(os.getcwd(), projectname))

    return project


@click.group()
@click.option("-v", "--verbose", is_flag=True)
@click.option("-m", "--mock", is_flag=True,
              help="Pypi server is test server, git won't push")
@click.option("-s", "--signed", is_flag=True,
              help="Check signatures whan DOWNLOADING packages")
@click.option("-r", "--repair", is_flag=True,
              help="Run autopep8 to make the code pep8 conform")
@click.option("-n", "--nocheck", is_flag=True,
              help="pep8 confirmity problem become non-fatal")
@click.option("-u", "--update", is_flag=True,
              help="Update dependencies before build")
@click.option("-t", "--notest", is_flag=True, help="Skip the test part")
@click.option("-y", "--yes", is_flag=True, help="No confirmation")
def cli(verbose, mock, signed, repair, nocheck, update, notest, yes):
    log.set_verbose(verbose)
    log.debug('pwd: ' + os.getcwd())

    cfg.config['mock'] = mock
    cfg.config['signed'] = signed
    cfg.config['repair'] = repair
    cfg.config['check'] = not nocheck
    cfg.config['update'] = update
    cfg.config['test'] = not notest
    cfg.config['ask'] = not yes
    if mock:
        log.warning('Mock Mode enabled')
    core.check_script_version()

@cli.command()
@click.argument('projectname', default=".")
def login(projectname):
    """ Save your login info to a crypted file """
    get_project(projectname).login()

@cli.command()
@click.argument('scriptname', default="default")
@click.argument('projectname', default=".")
def run(scriptname, projectname):
    """ Run a script defined in pyp.json """
    log.debug('Script from CLI: '+scriptname)
    get_project(projectname).run(scriptname)

@cli.command()
@click.argument('projectname', default=".")
def init(projectname):
    """ Initializes a spvm project, use init projectname to create sub-directory projectname"""
    log.debug('Running init')
    get_project(projectname).init()


@cli.command()
@click.option("-s", "--show", is_flag=True,
              help="Show the problems with the code")
@click.argument('projectname', default=".")
def status(projectname, show):
    """ Print information about the project """
    get_project(projectname).print_project_status(show)


@cli.command()
@click.argument('projectname', default=".")
def update(projectname):
    """ Update the dependencies """
    get_project(projectname).update_dependencies()


@cli.command()
@click.argument('projectname', default=".")
def test(projectname):
    """ Run the tests on the current project """
    get_project(projectname).run_test()


@cli.command()
@click.argument('dependency')
@click.argument('projectname', default=".")
def add(dependency, projectname):
    """ Add and install a dependency to the project """
    get_project(projectname).add_dependency(dependency)


@cli.command()
@click.argument('projectname', default=".")
def repair(projectname):
    """ Force pep8 compliance on project """
    get_project(projectname).repair()


@cli.command()
@click.argument('projectname', default=".")
def patch(projectname):
    """ Start pipeline for patch release (e.g 0.0.1 -> 0.0.2) """
    get_project(projectname).release('patch')


@cli.command()
@click.argument('projectname', default=".")
def major(projectname):
    """ Start pipeline for major release (e.g 0.4.8 -> 1.0.0) """
    get_project(projectname).release('major')


@cli.command()
@click.argument('projectname', default=".")
def minor(projectname):
    """ Start pipeline for minor release (e.g 0.0.73 -> 0.1.0) """
    get_project(projectname).release('minor')


@cli.command()
@click.argument('kind', default=".")
@click.argument('projectname', default=".")
def release(kind, projectname):
    """
    Test build and publish to PyPi and or docker
    The release kind can be:
    - pass: No versin increase
    - major, minior, patch
    - <number>: increase the version index by 1
    """
    get_project(projectname).release(kind)


# @cli.command()
# def blind():
#     """
#     pass


@cli.command()
@click.argument('targets', default="git,pypi,docker")
@click.argument('projectname', default=".")
def publish(targets, projectname):
    """
    Publish the project to the specified targets
    Namely, git, pypi and docker
    By defualt it wil try to publish to the 3
    I you only want a git and pypi publication for instance, use spvm publish git,pypi
    """
    tgts = targets.split(',')
    get_project(projectname).publish(
        git='git' in tgts,
        pypi='pypi' in tgts,
        docker='docker' in tgts)


@cli.command()
@click.argument('projectname', default=".")
def install(projectname):
    """ Install a setup.py """
    get_project(projectname).install_setup(True)


# @cli.command()
# @click.argument('projectname', default=".")
# def run():
#     """ Run scripts defined in .pypm """
#     pass


if __name__ == "__main__":
    cli()



import os
from os.path import join

metaFileName = "pyp.json"
scriptVersionCheckURL = None  # Last version
metaFileLocation = join(os.path.dirname(__file__), 'res', metaFileName)

if not os.path.isfile(metaFileLocation):
    raise EnvironmentError("Meta file template not found")

STATUS_PROJECT_DOES_NOT_EXIST = "does_not_exist"
STATUS_PROJECT_NOT_INITIALIZED = "not_initialized"
STATUS_PROJECT_INITIALIZED = "initialized"

PADLOCK = '🔒'
OPEN_PADLOCK = '🔓'

config = {
    'mock': False,
    'signed': False
    #FIXME add other config (see cmd)
}


class NoFailReadOnlyDict(object):
    def __init__(self, d, default=''):
        self.dict = d
        self.default = default

    def __getitem__(self, key):
        try:
            return self.dict[key]
        except KeyError:
            return self.default

    def __setitem__(self, key, value):
        raise AttributeError("This dictionary is read only")

    # def __getattribute__(self, name):
    #     raise AttributeError("Cannot access attributes")

    # def __setattr__(self, name, value):
    #     raise AttributeError("Cannot set attributes")



import splogger as log
from subprocess import Popen, PIPE, CalledProcessError
import sys
import os
import shutil
import pytest
import requests
from colorama import Fore
import hashlib
from os.path import join
import fnmatch
import gnupg
import json
import getpass
from .config import NoFailReadOnlyDict
from datetime import datetime

from . import config

FNULL = open(os.devnull, 'w')


def call_with_stdout(args, ignore_err=False,
                     stdout=PIPE, inp=None, stderr=PIPE):
    with Popen(args.split(' ') if type(args) == str else args, stdout=stdout, stdin=PIPE if inp is not None else None, stderr=stderr) as proc:
        out, err = proc.communicate(input=inp)
        if proc.poll() != 0 and not ignore_err:
            if log.get_verbose():
                log.error('Error from subprocess')
                if err is not None and err != '':
                    print('err: ' + str(err), file=sys.stderr)
                if out is not None and out != '':
                    print('out: ' + str(out), file=sys.stderr)
            raise CalledProcessError(proc.poll(), args, out, err)

        if log.get_verbose():
            log.debug('Output of '+repr(args))
            if out is not None:
                print(out.decode())
            if err is not None:
                print(err.decode())

        if out is not None:
            return out.decode()
@log.clear()
def read_logins():
    if os.path.isfile('.logins'):
        log.success(Fore.GREEN+config.PADLOCK+" Found crypted logins file"+Fore.RESET)
        gpg = gnupg.GPG()

        cr = None
        with open('.logins', 'r') as fh:
            cr = fh.read()
        
        crypt = None
        while True:
            passphrase = getpass.getpass('Passphrase for login file: ')
            crypt = gpg.decrypt(cr, passphrase = passphrase)

            if not crypt.ok:
                log.error(Fore.RED + config.PADLOCK + 'Could not unlock the logins file, is the passphrase correct?'+Fore.RESET)
                # raise ValueError('Could not uncrypt the login file')
            else:
                break

        cr = json.loads(crypt.data)
        log.success('Logins creation time: '+cr['creation_date'])
        log.success(Fore.GREEN+config.OPEN_PADLOCK+' Got logins for '+', '.join(cr['data'])+Fore.RESET)
        
        return NoFailReadOnlyDict(cr['data'], default = None)
    return None

def get_date(): 
    return datetime.now().strftime("%Y-%m-%d %H:%M:%S")

def ask_logins():
    if os.path.isfile('.logins'):
        log.warning(Fore.YELLOW+'The logins file already exist, overwrite it ?'+Fore.RESET)
        yes = input('Enter \'yes\' to overwrite: ')
        if yes != 'yes':
            return

    cr = {
        'creation_date': get_date(),
        'data': {}
    }

    print('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
    print(Fore.YELLOW+'\tLogins configuration\n'+Fore.RESET)
    print(Fore.GREEN+'You are going to be asked you credentials for this project\n'
          +'They will be stored under .logins crypted with AES and .logins append to .gitignore'+Fore.RESET)
          
    yes = input('Do you wish to continue? (Enter "yes" to continue): ')
    if yes != 'yes':
        log.error('Cancelled')
        return
    print('')

    def ask_login(arr, component):
        print(Fore.LIGHTGREEN_EX+'\nNow configuring login for: '+Fore.GREEN+component)
        print(Fore.RED+"Leave blank if Not Applicable"+Fore.RESET)
        login = input('Login: ')
        if login == '':
            return
        password = getpass.getpass()
        arr[component] = {'login': login, 'password': password}
    
    
    # print(Fore.RED+'Git login: '+Fore.CYAN+' Git login will no be asked, you are expected to use a credential helper for git '
    # +'if you wish to automatically push'+Fore.RESET)
    # print(Fore.GREEN+'This setup can launch the git credential helper for you <3'
    # +'\n'+Fore.YELLOW+'WARNING: The credentials are stored in clear text'+Fore.RESET)
    # yes = input('Save git credentials? (Enter "yes" to continue): ')
    # if yes == 'yes':
    #     call_git('config credential.helper store')
    #     call_git('push')

    ask_login(cr['data'], 'git')
    ask_login(cr['data'], 'pypi')
    ask_login(cr['data'], 'docker')

    gpg = gnupg.GPG()
    gpg.encrypt(json.dumps(cr), (), symmetric=True, output='.logins')

    try:
        call_git('check-ignore .logins')
    except CalledProcessError:
        with open('.gitignore', 'a') as fh:
            fh.write('\.logins')
        log.success('Appened .logins to .gitignore')

def call_python(module, args, stdout=None, stderr=None):
    mod = [] if module == '' else ['-m', module]
    return call_with_stdout(
        [sys.executable, *mod, *args.split(' ')], stdout=stdout, stderr=stderr)


def copy(a, b):
    assert os.path.isfile(a)
    with open(a, 'r') as ffh:
        with open(b, 'w+') as tfh:
            tfh.write(ffh.read())


def query_get(url, make_json=True):
    req = requests.get(url)
    if not req.ok:
        raise ValueError('Request Failed with code: ' + str(req.status_code))
    if make_json:
        return req.json()
    return req.content


def install_packages(args, check_signatures=None):
    if check_signatures is None:
        check_signatures = config.config['signed']

    if not check_signatures:
        call_pip('install ' + args, verbose=True)
        return

    piptmp = 'piptmp'

    @log.element('Download Packages', log_entry=True)
    def download():
        for pack in args.split(' '):
            log.set_additional_info(pack)
            call_pip('download -d ' + piptmp + ' ' + pack)

    @log.element('Checking Packages')
    def check_packages(base_url='https://pypi.python.org/pypi/'):
        log.fine('Checking packages in: ' + piptmp)
        unchecked = 0
        for f in os.listdir(piptmp):
            try:
                log.set_additional_info(f)
                f_ = piptmp + os.sep + f
                if not os.path.isfile(f_):
                    continue

                splited = f.split('-')
                log.debug('Checking ' + splited[0])
                package_info = query_get(base_url + splited[0] + '/' + splited[1] + '/json')

                for f_info in package_info['releases'][splited[1]]:
                    if not os.path.isfile(os.path.join(piptmp, f_info['filename'])):
                        continue

                    if md5(f_) != f_info['md5_digest']:
                        log.error('Hash do not match')
                        exit(1)
                    # log.success(Fore.GREEN+'Hash checked for '+f)

                    if not f_info['has_sig']:
                        log.debug(Fore.YELLOW + 'No signature provided for ' + f_info['filename'])  # FIXME throw?
                        unchecked += 1
                        continue

                    sig = query_get(f_info['url'] + '.asc', False)
                    log.debug('File: ' + f_info['filename'] + ' has signature:\n ' + sig.decode())

                    # Check
                    q = '' if log.get_verbose() else ' --quiet'
                    try:
                        call_gpg('--no-default-keyring --keyring tmp.gpg' + q + ' --auto-key-retrieve --verify - ' + f_, inp=sig)  # FIXME Only use known keys?
                    except CalledProcessError as er:
                        if er.returncode == 1:
                            log.error(Fore.RED + config.OPEN_PADLOCK + ' Invalid signature for ' + f)
                            exit(1)

                        log.error('Could not check signature for ' + f + ' (' + repr(er) + ')')
                        unchecked += 1
                        continue

                    log.success(Fore.GREEN + config.PADLOCK + ' File ' + f + ' is verified')

            except KeyboardInterrupt:
                exit(2)
            except SystemExit as e:
                raise e
            except BaseException as be:
                log.error(Fore.RED + config.OPEN_PADLOCK + ' Failed to check ' + f + Fore.RESET)
                log.error(repr(be))
        log.warning(Fore.YELLOW + str(unchecked) + ' file(s) could not be verified')

    def clearup():
        shutil.rmtree(piptmp, True)
        log.success('Cleaned temporary download directory')

    @log.element('Install Packages')
    def install():
        for f in os.listdir(piptmp):
            if f.endswith('.whl'):
                log.set_additional_info(f)
                call_pip('install ' + piptmp + os.path.sep + f)
                # log.success('Installed ' + f.split('-')[0])

    clearup()
    download()
    check_packages()
    install()
    clearup()

def match_gitignore(name, ignore):
    for pattern in ignore.split(','):
        if fnmatch.fnmatch(name, pattern):
            return True
    return False

def call_pip(args, verbose=log.get_verbose()):
    fh = PIPE if verbose else FNULL
    return call_python('pip', args, stdout=fh, stderr=fh)


@log.clear()
def call_pytest(args):
    o = pytest.main(args.split(' '))
    if o != 0:
        raise CalledProcessError(o, 'pytest ' + args)


@log.clear()
def call_git(args):
    if type(args) == str:
        args = 'git '+args
    else: # array
        args = ['git', *args]

    return call_with_stdout(args)


@log.element('Commiting', log_entry=True)
def call_commit(message, key=''):
    args = ['git', 'commit', '--no-edit']
    if key != '':
        args.append('-S' + key)
    args.append('-m')
    args.append(message)

    return call_with_stdout(args)


def md5(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def call_gpg(args, inp=None, verbose=log.get_verbose()):
    # log.error('Call to gpg deprecated')
    fh = PIPE if verbose else FNULL
    return call_with_stdout('gpg ' + args, inp=inp, stdout=fh, stderr=fh)


def call_twine(args):
    return call_with_stdout('twine ' + args, stdout=None) # FIXME import and use


@log.element('Checking code', log_entry=False)
def call_check(args, ignore="", exclude=''):
    flakes = ''
    for dirname, _, files in os.walk(args):
        if os.path.ismount(dirname) or os.path.islink(dirname):
            log.warning('Ignored mounted/link directory: '+dirname)
            continue

        if match_gitignore(dirname, exclude):
            continue

        for f in files:
            if not f.endswith('.py'):
                continue
            log.debug('Check: '+f)
            log.set_additional_info(f)
            if match_gitignore(join(dirname, f), exclude):
                continue 
            
            flak = call_with_stdout('python -m pyflakes ' + join(dirname, f), ignore_err=True)
            if flak is not None:
                flakes += flak

    # flakes = call_with_stdout('python -m pyflakes ' + args, ignore_err=True)
    pep8 = call_with_stdout(
        'python -m pycodestyle --ignore=' +
        ignore +     
        ' ' +
        ('' if exclude == '' else '--exclude='+exclude) +
        ' ' +
        args,
        ignore_err=True)
    return flakes, pep8


@log.clear()
def input_with_default(prompt, default=None, rtype=str):
    assert type(rtype) == type
    if default is None:
        default = rtype()

    while True:
        res = input(f'{prompt} ({default}): ')
        if res == 'null' or res == 'none':
            return ''
        if res is '':
            return default

        try:
            return rtype(res)
        except ValueError:
            log.error('An input of type ' + rtype.__name__ + " is expected")


__all__ = [
    'input_with_default',
    'call_check',
    'call_git',
    'call_pytest',
    'call_pip',
    'copy',
    'call_python',
    'call_with_stdout']



import splogger as log
import os.path
from os.path import join
import json
from colorama import Fore
import subprocess
from time import sleep
from shutil import rmtree
import urllib.parse
import docker
from threading import Thread
import re
from subprocess import CalledProcessError
import spvm

from . import config
from . import ioutils
from . import metautils


class PYVSProject(object):
    """
    A project object at the current cwd on which various actions are possible
    """

    def __init__(self, location=os.getcwd()):
        self.location = os.path.normpath(location)  # directory for project
        if os.path.isdir(self.location):
            os.chdir(self.location)
            log.debug("cwd switched: " + self.location)
        log.debug("New  project instance @ " + self.location)
        log.debug("Project status is: " + str(self.get_project_status()))
        self.projectMetaFile = join(self.location, config.metaFileName)
        self.maybe_load_meta()
        if self.meta is not None:
            metautils.check_project_meta(self.meta)
            self.save_project_info()
        self.load_gitignore()

    def init(self):
        """
        Create the spvm project folder and init a base project structure
        Guess project meta and asks for correction, write in metaFile
        """
        os.makedirs(join(self.location, 'test'), exist_ok=True)
        os.makedirs(join(self.location, os.path.basename(self.location).lower()), exist_ok=True)

        self.meta = metautils.detect_project_meta(self.location)
        # self.print_project_meta()
        while True:
            metautils.prompt_project_info(self.location, self.meta)
            self.print_project_meta()

            if metautils.input_with_default(f'{Fore.CYAN}Is this correct? {Fore.RESET}[y/n]') in ('y', 'Y', '1', 'yes', 'Yes'):
                break

        self.save_project_info()

        log.success('Project initialized')

    def print_project_meta(self):
        print(Fore.GREEN + '      Project Metadata' + Fore.RESET)
        print(json.dumps(self.meta, indent=4))

    def get_project_status(self):
        """
        Get the status, is the project initialized, not present?
        """

        if not os.path.isdir(self.location):
            # The project folder does not exist
            return config.STATUS_PROJECT_DOES_NOT_EXIST
        if not os.path.isfile(os.path.join(
                self.location, config.metaFileName)):
            # The project exists but the spvm project is not initiated
            return config.STATUS_PROJECT_NOT_INITIALIZED
        return config.STATUS_PROJECT_INITIALIZED        # The spvm project is initiated

    def check_code(self):
        """
        Run pyflakes and pycodestyle to check ode validity and pep8 conformity
        Reported problems are in the 2 returned arrays
        """

        flakes, pep = ioutils.call_check(self.location, ignore=self.meta['project_vcs']['ignored_errors'], exclude=self.gitignore)

        flakes = flakes.split('\n')[:-1]
        pep = pep.split('\n')[:-1]

        return flakes, pep

    def run_test(self):
        """
        Run the tests with pytest
        """
        try:
            ioutils.call_pytest(self.location)
        except CalledProcessError as ex:
            if ex.returncode == 5:
                log.warning('No tests were found')
                log.warning('This is not considered fatal but is VERY STRONGLY discouraged')
                log.warning('Resuming in 2 seconds')
                sleep(2)
                return
            log.error('Tests Failed')
            exit(1)
        log.success('Tests passed')

    def save_project_info(self):
        """
        Write the current project info to the metaFile
        """
        with open(self.projectMetaFile, 'w+') as fh:
            fh.write(json.dumps(self.meta, indent=4))
        log.debug('Saved project meta')

    @log.auto()
    def add_dependency(self, dep):
        """
        Add a dependency to the project:
        Pip handles different sources:
        - PyPI (and other indexes) using requirement specifiers.
        - VCS project urls.
        - Local project directories.
        - Local or remote source archives.
        """
        if self.get_project_status() != config.STATUS_PROJECT_INITIALIZED:
            log.error("The project is not initialized, call spvm init first before adding dependencies")
            return

        # if os.path.isdir(dep):
        #     log.warning(
        #         "Adding local dependency, this may break when releasing the project")

        # if os.path.isfile(dep):
        #     log.warning(
        #         "Got a file as dependency name, using as requirements.txt")
        #     supp_args = '-r '

        try:
            ioutils.install_packages(dep)
        except CalledProcessError as ex:
            log.error("Pip call failed with code " + str(ex.returncode))
            log.error("Does the dependency exist or is something broken?")
            return

        self.meta['project_requirements']['python_packages'].append(dep)
        self.save_project_info()

    def maybe_load_meta(self):
        """
        If a metaFile exists, it will be loaded
        """
        self.meta = None
        if os.path.isfile(self.projectMetaFile):
            try:
                with open(self.projectMetaFile, 'r') as fh:
                    self.meta = json.loads(fh.read())
                    log.debug('Loaded project meta')
            except json.JSONDecodeError as e:
                log.error(repr(e))
                log.error(config.metaFileName + " is invalid")
                exit(1)
        else:
            log.debug('No meta to load')

    def remove_dependency(self, dep):
        # TODO
        log.error('Not implemented: remove dependency')

    @log.element('Setup install')
    def install_setup(self, force=False):
        """
        Install the template setup.py
        """

        already_present = os.path.isfile(join(self.location, 'setup.py'))
        if already_present and not force:
            log.warning('setup.py already present, it will not be replaced')
            log.warning('Run spvm install to force setup.py replacement')
            return

        if already_present:
            log.fine('Creating setup.py.backup')
            ioutils.copy(join(self.location, 'setup.py'), join(self.location, 'setup.py.backup'))

        log.success('Copying seyup.py from template')
        ioutils.copy(join(os.path.dirname(__file__), 'res', 'setup.py'), join(self.location, 'setup.py'))

    @log.no_spinner()
    @log.element('Update project dependencies', log_entry=True)
    def update_dependencies(self):
        """
        Run a pip upgrade of the dependency list
        """
        # ioutils.call_pip(
        #     'install ' +
        #     " ".join(
        #         self.meta['project_requirements']['python_packages']) +
        #     " --upgrade")

        ioutils.install_packages(" ".join(self.meta['project_requirements']['python_packages']))

    def up_version(self, kind):  # FIXME other to 0
        """
        Increase the version in the project meta base on the 'kind' instruction:
        kind can be a str or a number
        if a number, it is treated like an index for the version increment
        if a string, it can be major, minor or patch
        """

        log.fine('Increasing version (' + str(kind) + ')')
        v = self.meta['project_vcs']['version']
        v_ = [int(i) for i in v.split('.')]
        while len(v_) < 3:
            v_.insert(0, '0')
        log.debug("Current version comphrension: " + str(v_))

        if kind.isdigit():
            kind = int(kind)
            if kind < 0 or kind >= len(v_):
                log.error('Unrecognized version changer: ' + str(kind))
            v_[int(kind)] += 1
        else:
            kind = kind.lower()
            if kind == 'patch':
                index = len(v_) - 1
            elif kind == 'major':
                index = 0
            elif kind == 'minor':
                index = len(v_) - 2
            elif kind == 'pass':
                log.success('Version not changed')
                return
            else:
                log.error('Unrecognized version changer: ' + str(kind))
                exit(1)

            v_[index] += 1
            v_ = [v_[i] if i <= index else 0 for i in range(len(v_))]

        self.meta['project_vcs']['version'] = '.'.join([str(i) for i in v_])
        self.save_project_info()
        log.success(v + ' -> ' + self.get_version())

    @log.element('🔧 Code repair', log_entry=True)
    def repair(self):
        ioutils.call_python('autopep8', '-ra --in-place .')

    def populate_init(self):
        """ Populate the <proj>/__init__.py with meta info """
        log.debug('Populating the __init__')
        init_path = join(self.get_name().lower(), '__init__.py')

        with open(init_path, 'r') as fh:
            init_file = fh.read()

        def replace_or_create(key, value):
            nonlocal init_file
            pattern = "^" + key + " ?= ?.*"  # key at the beginning of the line
            preg = re.compile(pattern, re.M)

            if len(preg.findall(init_file)) == 0:
                init_file += '\n' + key + ' = "' + value + '"'
                log.debug(key + ' not found in init')
            else:
                init_file = preg.sub(key + ' = "' + value + '"', init_file)
                log.debug('Found ' + key + ' and replaced with ' + value)

        replace_or_create('__name__', self.meta['project_info']['name'])
        replace_or_create('__version__', self.meta['project_vcs']['version'])
        replace_or_create('__author__', self.meta['project_authors'][0]['name'])
        replace_or_create('__url__', self.meta['project_info']['url'])
        replace_or_create('__email__', self.meta['project_authors'][0]['email'])

        os.remove(init_path)
        with open(init_path, 'w+') as fh:
            fh.write(init_file)

        log.success('Populated __init__.py')

    # RELEASE #

    def detect_publish_context(self):
        """
        Indicates where the code will be pushed
        Inspects the config map and project's meta
        """

        context = [False,  # Git
                   False,  # pypi
                   False]  # docker

        mock = config.config['mock']
        has_pypi_rep = self.meta['project_vcs']['pypi_repository'] != ''
        has_docker_rep = self.meta['project_vcs']['docker_repository'] != ''
        has_dockerfile = os.path.isfile(join(self.location, 'Dockerfile'))

        if not mock:
            context[0] = True

        if has_pypi_rep:
            context[1] = True

        if has_docker_rep and has_dockerfile:
            context[2] = True

        return context

    def release(self, kind='pass'):
        """
        Starts a release pipeline
        """

        if self.get_project_status() != config.STATUS_PROJECT_INITIALIZED:
            log.error('The project is not initialized')
            log.error('Run spvm init first')
            exit(1)

        pipeline = []
        log.fine('Calculating release pipeline')

        pipeline.append(self.clear_build)
        if config.config['update']:
            pipeline.append(self.update_dependencies)
        if config.config['repair']:
            pipeline.append(self.repair)
        pipeline.append(self.check_project)
        if config.config['test']:
            pipeline.append(self.run_test)
        pipeline.append(self.up_version)
        pipeline.append(self.populate_init)
        pipeline.append(self.install_setup)
        pipeline.append(self.publish)

        NO = Fore.RED + 'NO' + Fore.RESET
        MOCK = (
            '' if not config.config['mock'] else ' ' +
            Fore.LIGHTYELLOW_EX +
            '(MOCK)' +
            Fore.RESET)
        YES = Fore.GREEN + 'YES' + Fore.RESET + MOCK

        publish_context = self.detect_publish_context()
        pipeline.append(Fore.CYAN + "     - Git Publish:\t\t"    + (YES if publish_context[0] else NO))
        pipeline.append(Fore.CYAN + "     - PyPi Publish:\t\t"   + (YES if publish_context[1] else NO))
        pipeline.append(Fore.CYAN + "     - Docker Publish:\t"   + (YES if publish_context[2] else NO))

        log.success('Release pipeline is: ')
        for f in pipeline:
            if isinstance(f, str):
                log.success(f)
                continue
            log.success(" -> " + f.__name__)

        if not config.config['mock']:
            log.warning(
                Fore.YELLOW +
                'The mock mode is not activated, this is for real !' +
                Fore.RESET)

        if config.config['ask']:
            input('Press Enter to continue')

        for f in pipeline:
            if isinstance(f, str) or f.__name__ == 'wrapper':
                continue

            log.success('> ' + f.__name__)
            if f.__name__ == 'up_version':  # the only one to give parameters to
                f(kind)
            else:
                f.__call__()

    def check_project(self):
        """ Check code and exit if not conform """
        pf, pe = self.check_code()

        if len(pf) + len(pe) > 0:
            log.error('Project is not conform or has errors, run spvm status -s')
            if config.config['check']:
                log.error(
                    Fore.RED +
                    'This error is fatal, to make it non-fatal, use -n')
                exit(1)

    @log.element('Building', log_entry=True)
    def build(self):
        log.success('Building package in ./build')
        try:
            ioutils.call_python('', 'setup.py sdist -d build/dist bdist_wheel -d build/dist', stdout=subprocess.PIPE)
        except CalledProcessError as ex:
            log.error('Unable to build the package')
            log.error(repr(ex))
            exit(1)

    @log.element('Cleaning up', log_entry=True)
    def clear_build(self):
        # remove ./build ./<name>.egg-info
        rmtree('./build', True)
        rmtree('./' + self.get_name() + '.egg-info', True)

    @log.element('Publishing', log_entry=True)
    def publish(self, git=True, pypi=True, docker=True):
        context = self.detect_publish_context()
        git = git and context[0]
        pypi = pypi and context[1]
        docker = docker and context[2]

        logins = ioutils.read_logins()

        if git and not config.config['mock']:
            self._release_git(credentials = logins['git'])
        if pypi:
            self._release_pypi(credentials = logins['pypi'])
        if docker:
            self._release_docker(credentials = logins['docker'])

    @log.element('Git Publishing', log_entry=True)
    def _release_git(self, credentials = None):
        if os.path.isfile('.git-credentials'):
            os.remove('.git-credentials')
            log.success('Removed dangling credential file')


        # Commit version
        commit_message = self.meta['project_vcs']['release']['commit_template'].replace('%s', self.meta['project_vcs']['version']).replace('"', '\\"').strip()
        log.debug('Commit message: ' + commit_message)
        ioutils.call_git('add .')

        key = self.meta['project_vcs']['release']['git_signing_key']
        if key != '':
            log.success(Fore.GREEN + config.PADLOCK + 'Commit will be signed with ' + key)

        ioutils.call_commit(commit_message, key=key)

        # Tag version
        tag = self.meta['project_vcs']['release']['tag_template'].replace('%s', self.meta['project_vcs']['version'])
        ioutils.call_git('tag ' + ('' if key == '' else '-u ' + key + ' ') + '-m ' + tag + ' ' + tag)
        log.success('Tagged: ' + tag)

        try:

            # Login
            if credentials != None:
                log.fine('Setting git credentials to temporary file')    
                u = urllib.parse.urlparse(self.meta['project_vcs']['code_repository'])
                with open('.git-credentials', 'w+') as fh:
                    fh.write(u.scheme+'://'+credentials['login']+':'+credentials['password']+'@'+u.hostname+'\n')
                ioutils.call_git(['config', 'credential.helper', 'store --file .git-credentials', '--replace-all'])
                log.success('Credentials are set')

            # Push
            repo = self.meta['project_vcs']['code_repository']
            log.success('Pushing to ' + repo)
            ioutils.call_git('push ' + repo + ' --signed=if-asked')
            log.success('Pushing tags')
            ioutils.call_git('push ' + repo + ' --tags --signed=if-asked')

        finally:
            if credentials != None:
                os.remove('.git-credentials')
                log.success('Removed temporary credential file')

    
    @log.element('Package Release', log_entry=True)
    def _release_pypi(self, sign=True, credentials = None):
        # 🔒 🔐 🔏 🔓
        if self.meta['project_vcs']['pypi_repository'] == '':
            log.success('Nothing to push to pypi')
            return
        self.clear_build()
        self.build()
        if sign:
            self._sign_package()
        self._pypi_upload(credentials)
        self.clear_build()

    @log.clear()
    def _pypi_upload(self, credentials = None):
        mock = config.config['mock']

        # Upload
        if mock:
            rep = "https://test.pypi.org/legacy/"  # FIXME put in config
        else:
            rep = self.meta['project_vcs']['pypi_repository']
        log.success('Uploading to ' + rep)
        ioutils.call_twine('upload --repository-url ' + rep + ' ./build/dist/*' + ('' if credentials == None else ' -u '+credentials['login']+ ' -p '+credentials['password']))

    @log.element('Package Signing')
    def _sign_package(self):
        """
        Add the signatures to the package before upload
        """
        meta_key = self.meta['project_vcs']['release']['package_signing_key']
        if meta_key == '':
            log.error(Fore.RED + config.OPEN_PADLOCK + ' No key provided for package signing' + Fore.RESET)
            return

        log.success('Signing the package with the key: ' + meta_key)

        try:
            for place in os.walk(join('.', 'build', 'dist')):
                for f in place[2]:
                    self._sign_file(join(place[0], f), meta_key)
        except CalledProcessError as ex:
            log.error(Fore.RED + config.OPEN_PADLOCK + ' Could not sign the package' + Fore.RESET)
            log.error('The program will now stop, you can resume with: spvm publish pypi')
            log.error('When the issues are fixed')
            log.error(repr(ex))
            exit(1)
            return

        log.success(Fore.GREEN + config.PADLOCK + ' Package Signed with key: ' + meta_key + Fore.RESET)

    def _sign_file(self, file, meta_key):
        ioutils.call_gpg(('-u ' + meta_key + ' ' if meta_key != '' else '') + '-b --yes -a -o ' + file + '.asc ' + file)
    
    @log.element('Docker Publishing', log_entry=True)
    def _release_docker(self, credentials = None):
        log.success('Building Docker Image')
        client = docker.from_env()
        log.debug(json.dumps(client.version(), indent=4))

        status = {}

        def _show_docker_progress(obj):
            nonlocal status

            if 'errorDetail' in obj:
                log.error(Fore.RED + 'Error: ' + str(obj['errorDetail']['message']) + Fore.RESET)
                raise docker.errors.DockerException(obj['errorDetail']['message'])

            if 'stream' in obj:
                for line in obj['stream'].split('\n'):
                    if line == '':
                        continue
                    log.success(line.strip())
                status.clear()
                return

            if 'status' in obj:
                if 'id' not in obj:
                    log.success(obj['status'])
                    return

                if len(status) == 0:
                    print(Fore.GREEN + "\rA docker I/O operation is in progress" + Fore.RESET)

                s = obj['id'].strip() + ' ' + obj['status'] + '\t'
                if 'progress' in obj:
                    s += obj['progress']

                if obj['id'] not in status:
                    status[obj['id']] = {'index': len(status) + 1, 'str': s}
                    print(s)
                    return

                status[obj['id']]['str'] = s
                print('\033[F' * (len(status) + 1))
                for e in status:
                    print('\033[K' + status[e]['str'])
                # print('\n'*i, end = '')

        # FIXME choose dockerfile
        rep = self.meta['project_vcs']['docker_repository']
        log.success('Image repo: ' + rep)
        if hasattr(client, 'api'):
            client = client.api

        g = client.build(tag=rep, path='.', dockerfile='Dockerfile')

        for line in g:
            _show_docker_progress(json.loads(line.decode()))

        if config.config['mock']:
            log.warning(Fore.YELLOW + 'Mock mode: not pushing' + Fore.RESET)
            return

        if credentials != None:
            try:
                client.login(credentials['login'], credentials['password'])
            except docker.errors.APIError as excep:
                log.error('Cannot login: '+str(excep))
                exit(1)
            log.success('Logged in as '+credentials['login'])

        log.success('Pushing image')
        for line in client.push(rep, stream=True):
            _show_docker_progress(json.loads(line.decode()))

    def run(self, scriptname):
        if scriptname not in self.meta['scripts']:
            log.error('Script not found: '+scriptname)
            exit(1)
        
        self._run_scripts_pipeline(scriptname)

    @log.clear()
    def _run_scripts_pipeline(self, name):
        if 'pre'+name in self.meta['scripts']:
            self._run('pre'+name)
        self._run(name)
        if 'post'+name in self.meta['scripts']:
            self._run('post'+name)
        

    def _run(self, name):
        log.success('Running script: '+name)
        script = self.meta['scripts'][name]
        ioutils.call_with_stdout(['/bin/sh', '-c', script], stdout=None, stderr=None)


    def login(self):
        ioutils.ask_logins()

        

    # PRINT INFOS #

    def print_version_status(self):
        """
        Print information about the version of the project
        And code versioning related
        """
        p = nice_print_value

        p('     Version Info', kc=Fore.GREEN)
        if not os.path.isdir(join(self.location, '.git')) or ioutils.call_git('branch') == '':
            print(Fore.RED + 'No git repo initialized, versioning info not available' + Fore.RESET)
            p('')
            return

        tags = ioutils.call_git('tag --sort=-creatordate') + '\n' + 'None'

        p('Current Version:', self.get_version())
        p('Current branch:', ioutils.call_git('branch --no-color').split()[1])
        p('Current commit:', ioutils.call_git('rev-parse --short HEAD').strip())
        p('Commit count:', ioutils.call_git('rev-list --all --count').strip())
        p('Last Tag:', tags.split()[0])
        p('')

    def print_code_status(self, show=False):
        flakes, pep = self.check_code()
        p = nice_print_value

        p('\r     Code Status   ', kc=Fore.GREEN)

        vcf = Fore.LIGHTGREEN_EX if len(flakes) == 0 else Fore.LIGHTRED_EX
        vcp = Fore.LIGHTGREEN_EX if len(pep) == 0 else Fore.LIGHTRED_EX

        p('Errors & Warnings:', str(len(flakes)), vc=vcf)
        p('Conformity Problems:', str(len(pep)), vc=vcp)
        p('')

        if show:
            p('Errors & Warnings', kc=Fore.LIGHTYELLOW_EX)
            if len(flakes) == 0:
                print(Fore.GREEN + ' Nothing to display')
            for ew in flakes:
                print('> ' + Fore.RED + ew + Fore.RESET)
            p('')
            p('Conformity issues', kc=Fore.LIGHTYELLOW_EX)
            if len(pep) == 0:
                print(Fore.GREEN + ' Nothing to display')
            for ew in pep:
                print('> ' + Fore.RED + ew + Fore.RESET)

    def print_dependencies(self):
        p = nice_print_value

        p('     Project Dependencies', '(' + str(len(self.meta['project_requirements']['python_packages'])) + ')', kc=Fore.GREEN)
        for name in self.meta['project_requirements']['python_packages']:
            print(name)

        p('')

    @log.element(action='Calculating size...')
    def get_project_size(self):
        total_size = 0
        for dirpath, _, filenames in os.walk(self.location):

            if self.match_gitignore(dirpath):
                log.debug('Ignored ' + dirpath)
                continue

            for f in filenames:
                fp = os.path.join(dirpath, f)
                total_size += os.path.getsize(fp)

        return sizeof_fmt(total_size)

    def match_gitignore(self, name):
        return ioutils.match_gitignore(name, self.gitignore)

    def load_gitignore(self):
        if not os.path.isfile(join(self.location, '.gitignore')):
            self.gitignore = ''
            return

        with open(join(self.location, '.gitignore'), 'r') as fh:
            gi = fh.read()
            gi = gi.replace('\n\n', '\n').strip().split('\n')
            gi = [join(self.location, e) for e in gi]
            gi.append(join(self.location, '.git'))
            self.gitignore = ','.join(gi)
            log.debug('Ignoring: ' + self.gitignore)

    @log.clear()
    def print_project_status(self, show=False):
        """
        Print information about the project
        """
        p = nice_print_value
        sts = self.get_project_status()

        if sts != config.STATUS_PROJECT_INITIALIZED:
            print(Fore.RED + 'Enable SPVM to print project status\nspvm init' + Fore.RESET)
            return

        nice_print_value('      Project Info', kc=Fore.GREEN)
        p('~~~~~~~~~~~~~~~~~~~~~~~~~~')
        p('Project Name:', self.get_name())
        p('Project Version:', self.get_version())
        p('Project Size:', self.get_project_size())
        p('Project SPVM Status:', sts)
        p('')
        p('~~~~~~~~~~~~~~~~~~~~~~~~~~')

        self.print_version_status()
        p('~~~~~~~~~~~~~~~~~~~~~~~~~~')
        self.print_code_status(show)
        p('~~~~~~~~~~~~~~~~~~~~~~~~~~')
        self.print_dependencies()
        p('~~~~~~~~~~~~~~~~~~~~~~~~~~')
        # self.print_project_meta()
        # print_docker_status()
        # print_

    # Userfull getters

    def get_version(self):
        return self.meta['project_vcs']['version']

    def get_name(self):
        return self.meta['project_info']['name']


def nice_print_value(key, value='', kc=Fore.WHITE, vc=Fore.LIGHTBLUE_EX):
    print(f'{kc}{key} {vc}{value}{Fore.RESET}\033[K')


def make_project_object(location):
    return PYVSProject(location)


def sizeof_fmt(num, suffix='B'):
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)

# version checking


def check_script_version():

    def _retreive_version():
        try:
            version = spvm.__version__
            log.debug("SPVM version " + str(version))
            lastver = ioutils.query_get("https://pypi.org/pypi/spvm/json")['info']['version']
            log.debug('Last version is ' + lastver)
            if version != lastver:
                log.warning("A new version of spvm is available (" + lastver + ") you have version " + version)
                log.warning("Run pip install spvm --upgrade")
        except BaseException:
            log.warning('Could not get last version')

    Thread(target=_retreive_version, daemon=True).start()



import splogger as log
from .config import metaFileName, NoFailReadOnlyDict, metaFileLocation
from .ioutils import input_with_default, call_git
from colorama import Fore
import getpass
import os
import json
from os.path import join


def get_default_template():
    """
    Get the default metaFile to be used
    """
    with open(metaFileLocation, 'r') as f:
        template = f.read()
    return json.loads(template)


@log.element('Check meta')
def check_project_meta(meta):
    """
    Check the given meta so it matchs the template, adding  elements
    """

    def rec_check_meta(meta_elem, template_elem):
        if type(meta_elem) not in [dict, list]:
            return
        for t_elem in template_elem:
            if t_elem not in meta_elem:
                meta_elem[t_elem] = template_elem[t_elem]
                log.warning('Added ' + t_elem + ' to your project meta with default value')
            else:
                if isinstance(template_elem[t_elem], type({})):
                    rec_check_meta(meta_elem[t_elem], template_elem[t_elem])
                if isinstance(template_elem[t_elem], type([])):
                    if len(template_elem[t_elem]) > 0:
                        for e in meta_elem[t_elem]:
                            rec_check_meta(e, template_elem[t_elem][0])

    rec_check_meta(meta, get_default_template())
    log.debug('Checked project meta')


@log.auto()
def detect_project_meta(location):
    assert os.path.isdir(location)
    meta = get_default_template()

    # detect metaFile
    if os.path.isfile(join(location, metaFileName)):
        log.success('Found ' + metaFileName)
        with open(join(location, metaFileName), 'r') as f:
            lines = f.read()
        return json.loads(lines)

    # detect setup.py
    if os.path.isfile(join(location, 'setup.py')):
        log.success("Found setup.py")
        setup_info = detect_from_setup(join(location, 'setup.py'))
        log.debug('Setup detection yield: ' + setup_info)

        meta['project_info']['name'] = setup_info['name']
        meta['project_info']['description'] = setup_info['description']
        meta['project_info']['license'] = setup_info['license']
        meta['project_info']['url'] = setup_info['url']

        meta['project_authors'][0]['name'] = setup_info['author']
        meta['project_authors'][0]['email'] = setup_info['author_email']

        meta['project_vcs']['version'] = setup_info['version']

        meta['project_requirements']['python_packages'] = setup_info['install_requires']
        meta['project_requirements']['python_version'] = setup_info['python_requires']

    # detect .git
    if os.path.isdir(join(location, '.git')):
        log.success("Found git structure")
        remote = call_git('remote').split(' ')[0].strip()
        if remote != '':
            log.fine("Found remote " + remote)
            meta['project_vcs']['code_repository'] = call_git(
                'remote get-url ' + remote).strip()

        if meta['project_authors'][0]['email'] == '':
            log.fine('Using git to detect email')
            meta['project_authors'][0]['email'] = call_git(
                'config user.email').split()[0]

    # Find project name if not detected
    if meta['project_info']['name'] == '':
        meta['project_info']['name'] = os.path.basename(location)
        log.fine('Using default name: ' + meta['project_info']['name'])

    # detect __version__  (sort of common declaration)
    if meta['project_vcs']['version'] == '':
        meta['project_vcs']['version'] = load_version_from_file(
            location, meta['project_info']['name'])

    # other methods
    if meta['project_authors'][0]['name'] == '':
        meta['project_authors'][0]['name'] = getpass.getuser()
        log.fine('Using username as project author')

    if meta['project_vcs']['code_repository'] == '':
        log.fine('Using github private repo as code_repo')
        meta['project_vcs']['code_repository'] = "https://github.com/" + \
            getpass.getuser() + "/" + meta['project_info']['name']

    return meta


@log.clear()
def prompt_project_info(location, meta=get_default_template()):
    """
    Prompt the user for project information
    """

    print(f'\n{Fore.GREEN}~~~~ Project Setup ~~~~{Fore.RESET}')
    print(f'{Fore.LIGHTGREEN_EX}Project Location: {location}{Fore.RESET}')
    print(f'Enter {Fore.RED}null{Fore.RESET} or {Fore.RED}none{Fore.RESET} to leave a field blank\n')

    print(f'{Fore.CYAN}1/4:{Fore.LIGHTBLUE_EX} Project Info {Fore.RESET}')
    meta['project_info']['name'] = input_with_default('Project Name', meta['project_info']['name'])
    meta['project_info']['description'] = input_with_default('Project Description', meta['project_info']['description'])
    meta['project_info']['license'] = input_with_default('Project License', meta['project_info']['license'])
    meta['project_info']['url'] = input_with_default('Project URL', meta['project_info']['url'])

    print(f'{Fore.CYAN}2/4:{Fore.LIGHTBLUE_EX} Project Author {Fore.RESET}')
    meta['project_authors'][0]['name'] = input_with_default('Name', meta['project_authors'][0]['name'])
    meta['project_authors'][0]['url'] = input_with_default('URL', meta['project_authors'][0]['url'])
    meta['project_authors'][0]['email'] = input_with_default('Email', meta['project_authors'][0]['email'])
    print(f'{Fore.LIGHTGREEN_EX}Note that you can add more authors in the project {metaFileName} later{Fore.RESET}')

    print(f'{Fore.CYAN}3/4:{Fore.LIGHTBLUE_EX} Project Version Control {Fore.RESET}')
    meta['project_vcs']['code_repository'] = input_with_default('Code Repo', meta['project_vcs']['code_repository'])
    print(f'{Fore.GREEN}SPVM does not upload to Docker by default, add a repository to upload to\nand spvm will generate a generic Dockerfile{Fore.RESET}')
    meta['project_vcs']['docker_repository'] = input_with_default('Docker Repo', meta['project_vcs']['docker_repository'])  # f'{getpass.getuser()}/{meta.project_info.name}'
    print(f'{Fore.GREEN}SPVM does not upload to PyPi by default use: {Fore.CYAN}https://upload.pypi.org/legacy/{Fore.GREEN} as a pypi_repo to enable it (or use your own ;) ){Fore.RESET}')
    meta['project_vcs']['pypi_repository'] = input_with_default('PyPi Repo', meta['project_vcs']['pypi_repository'])
    meta['project_vcs']['version'] = input_with_default('Current Version', meta['project_vcs']['version'])

    print(f'{Fore.CYAN}4/4:{Fore.LIGHTBLUE_EX}Project Requirements{Fore.RESET}')
    meta['project_requirements']['python_version'] = input_with_default('Python Version', meta['project_requirements']['python_version'])
    dep = meta['project_requirements']['python_packages']
    print(f'{Fore.GREEN}Current dependencies: {" ".join(dep)}{Fore.RESET}')
    print(f'{Fore.GREEN}Add 1 dependency at a time and press Enter\nLeave blank to finish{Fore.RESET}')

    while True:
        tmp = input_with_default('Dependency', '')
        if tmp == '':
            break
        dep.append(tmp)
    meta['project_requirements']['python_packages'] = dep

    return meta


@log.auto()
def detect_from_setup(setup_file):
    setup_info = {}

    def fake_setup(**kwargs):
        nonlocal setup_info
        setup_info = kwargs

    import setuptools
    original_setup = setuptools.setup
    setuptools.setup = fake_setup

    try:
        with open(setup_file, 'r') as fh:
            setup_code = fh.read()

        exec(setup_code, globals())
        return NoFailReadOnlyDict(setup_info)
    except BaseException:
        log.error('Could not detect project config from setup.py')
        return None
    finally:
        setuptools.setup = original_setup


@log.auto()
def load_version_from_file(location, project_name):
    default_version = ''
    try:
        versionfile = join(location, project_name, '__version__.py')
        if os.path.isfile(versionfile):
            loaded_vfile = {}
            with open(versionfile, 'r') as vfile:
                content = vfile.read()
            exec(content, loaded_vfile)
            if loaded_vfile['__version__']:
                default_version = loaded_vfile['__version__']
                log.success("Found version " + default_version + " for project " + project_name)
    except BaseException as ex:
        log.error('Could not load file version: ' + ex.__class__.__name__)

    return default_version


__all__ = [
    'load_version_from_file',
    'detect_from_setup',
    'prompt_project_info',
    'detect_project_meta',
    'get_default_template']
